import numpy as np
from scipy import special

from rtdpy.rtd import RTD, RTDInputError


class Ncstr(RTD):
    """
    Create N CSTR in series (N-CSTR) AKA Tank in Series
    Residence Time Distribution (RTD) model. [1]_

    .. math::

        E(t) = \\frac{1}{\\tau}\\left(\\frac{t}{\\tau}\\right)^{n-1}
        \\frac{n^n}{\\left(n-1\\right)!}
        \\text{exp}\\left[\\frac{-nt}{\\tau}\\right]

    Parameters
    ----------
    tau : scalar
        Mean residence time of **all** CSTRs. ``tau>0``
    n : scalar
        Number of CSTRs. Can be a real number ``n>0``
    dt : scalar
        Time step for RTD. ``dt>0``
    time_end : scalar
        End time for RTD. ``time_end>0``

    References
    ----------
    .. [1] Levenspiel O. (1999) "Chemical Reaction Engineering: Third Edition"
           John Wiley & Sons, Inc.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import rtdpy
    >>> for n in [1, 2, 10]:
    ...     a = rtdpy.Ncstr(tau=1, n=n, dt=.01, time_end=3)
    ...     plt.plot(a.time, a.exitage, label="n={}".format(n))
    >>> plt.xlabel('Time')
    >>> plt.ylabel('Exit Age Function')
    >>> plt.legend()
    >>> plt.show()
    """

    def __init__(self, n, tau, dt, time_end):
        super().__init__(dt, time_end)
        self._n = None
        self._tau = None

        self.n = n
        self.tau = tau

    def _calc_exitage(self):
        """calculte exit age function"""
        try:
            output = (self.time / self.tau) ** (self.n - 1) / self.tau \
                * self.n ** self.n / special.gamma(self.n) \
                * np.exp(-self.time * self.n / self.tau)
        except (AttributeError, TypeError):
            output = None
        return output

    @property
    def n(self):
        """Number of CSTRS in series"""
        return self._n

    @n.setter
    def n(self, n):
        if n <= 0:
            raise RTDInputError("n less than zero")
        self._n = n
        self._exitage = self._calc_exitage()

    @property
    def tau(self):
        """Mean Residence Time of **all** tanks combined"""
        return self._tau

    @tau.setter
    def tau(self, tau):
        if tau <= 0:
            raise RTDInputError("tau less than zero")
        self._tau = tau
        self._exitage = self._calc_exitage()



import numpy as np
from scipy.integrate import cumtrapz

from rtdpy.const import DTTOL


class RTDInputError(Exception):
    pass


class RTD:
    """
    Base class for RTD analysis

    Child classes must implement:
        exitage
            The exitage function evaluated at each time point
    """
    def __init__(self, dt, time_end):
        self._dt = None
        self._time_end = None

        self._exitage = None
        self.dt = dt
        self.time_end = time_end

    @property
    def exitage(self):
        """Exit age distribution for RTD"""
        return self._exitage

    @property
    def exitage_norm(self):
        """Normalized Exit Age Distribtion for RTD"""
        return self.exitage / self.integral()

    @property
    def stepresponse(self):
        """Step respose of RTD"""
        return cumtrapz(self.exitage, self.time, initial=0)

    @property
    def stepresponse_norm(self):
        """Step respose of RTD"""
        return cumtrapz(self.exitage_norm, self.time, initial=0)

    @property
    def dt(self):
        """Time step for RTD"""
        return self._dt

    @dt.setter
    def dt(self, dt):
        if dt <= 0:
            raise RTDInputError
        self._dt = dt

    @property
    def time_end(self):
        """Last time point for RTD"""
        return self._time_end

    @time_end.setter
    def time_end(self, time_end):
        if time_end is None:
            self._time_end = None
            return
        if time_end <= 0:
            raise RTDInputError
        self._time_end = time_end

    @property
    def time(self):
        """
        Time points for exitage function
        """
        return np.arange(0, self.time_end, self.dt)

    def integral(self):
        """
        Integral of RTD
        """
        return np.trapz(self.exitage, self.time)

    def mrt(self):
        """
        Mean residence time of RTD
        """
        return np.trapz(self.exitage * self.time, self.time)

    def sigma(self):
        """
        Variance of RTD
        """
        tbar = self.mrt()
        return np.trapz(self.exitage*(self.time - tbar)**2, self.time)

    def output(self, inputtime, inputsignal):
        """
        Convolves input signal with RTD

        Parameters
        ----------
        inputtime : ndarray
            Times of input signal, which must have same `dt` as RTD.  Size m
        inputsignal : ndarray
            Input signal. Size n

        Returns
        -------
        outputsignal : ndarrary
            Output signal at same `dt`.  Size m + n -1
        """
        dt = np.diff(inputtime)
        udt = np.unique(dt)
        adt = np.average(dt)

        if (np.amax(udt) - np.amin(udt)) / adt > DTTOL:
            raise RTDInputError('input signal time does not have constant dt')

        if (adt-self.dt) / self.dt > DTTOL:
            print('time dt:' + str(adt))
            print('rtd dt:' + str(self.dt))
            raise RTDInputError('Cinput signal time dt does not match RTD dt')

        return np.convolve(inputsignal, self.exitage, mode='full') * self.dt

    def funnelplot(self, times, disturbances):
        """
        Return maximum output signal due to square disturbances.

        Also returns meshgrid for times and disturbance inputs
        for ease of plotting.

        Parameters
        ----------
        times : array_like, size m
            Times to determine funnelplot
        disturbances : array_like, size n
            Disturbance magnitudes

        Returns
        -------
        x : 2D meshgrid size (mxn)
            times
        y : 2D meshgrid size (mxn)
            disturbances
        response : 2D meshgrid size (mxn)
            maximum response at (x,y)
        """
        n = times.size
        m = disturbances.size

        response = np.zeros((m, n))
        for i, itime in enumerate(times):
            for j, jdisturbance in enumerate(disturbances):

                # input signal is a square disturbance of magnitude
                # jdisturbance and length itime
                inputsignal = np.zeros(self. time.size)
                inputsignal[self.time < itime] = jdisturbance

                # generate time stamps for generating output signal from RTD
                t = np.arange(0, self.time.size) * self.dt
                tmp = self.output(t, inputsignal)

                # store whichever deviation is largest
                if abs(np.amax(tmp)) > abs(np.amin(tmp)):
                    response[j, i] = np.amax(tmp)
                else:
                    response[j, i] = np.amin(tmp)

        x, y = np.meshgrid(times, disturbances)
        return x, y, response

    def frequencyresponse(self, omegas):
        """
        Parameters
        ----------
        omegas : ndarray
            frequencies at which to evaluate magnitude response

        Returns
        -------
        magnitude : ndarray
            frequency magnitude response at omegas
        """

        exitage = self.exitage
        series = np.arange(exitage.size)
        dt = self.dt
        mag = np.zeros(omegas.size)
        for i, omega in enumerate(omegas):
            mag[i] = dt * np.abs(
                np.sum(exitage * np.exp(-1 * 1j * series * omega * dt)))

        return mag



import numpy as np

from rtdpy.rtd import RTD, RTDInputError


class Pfr(RTD):
    """
    Create Plug Flow Reactor (PFR) Residence Time Distribution (RTD) model.
    [1]_

    .. math::

        E(t) = \\delta\\left(t-\\tau\\right)

    Parameters
    ----------
    tau : scalar
        Mean residence time of PFR. ``tau>0``
    dt : scalar
        Time step for RTD. ``dt>0``
    time_end : scalar
        End time for RTD. ``time_end>0``

    References
    ----------
    .. [1] Levenspiel O. (1999) "Chemical Reaction Engineering: Third Edition"
           John Wiley & Sons, Inc.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import rtdpy
    >>> a = rtdpy.Pfr(tau=1, dt=.01, time_end=3)
    >>> plt.plot(a.time, a.exitage)
    >>> plt.xlabel('Time')
    >>> plt.ylabel('Exit Age Function')
    >>> plt.show()
    """
    def __init__(self, tau, dt, time_end):
        super().__init__(dt, time_end)
        self._tau = None
        self._exitage = None

        self.tau = tau

    def _calc_exitage(self):
        try:
            t = self.time
            if self.tau >= self.time[-1]:
                raise RTDInputError(
                    'PFR lag is at or beyond end of supplied time')

            a = np.zeros(t.size)
            idx = np.nonzero(t <= self.tau)
            idxfirst = idx[0][-1]
            if idxfirst == t.size:
                raise RTDInputError('PFR lag is at end of supplied time')

            a[idxfirst] = (t[idxfirst + 1] - self.tau) / self.dt / self.dt
            a[idxfirst + 1] = (self.tau - t[idxfirst]) / self.dt / self.dt

            output = a
        except AttributeError:
            output = None
        return output

    @property
    def tau(self):
        return self._tau

    @tau.setter
    def tau(self, tau):
        if tau <= 0:
            raise RTDInputError("tau less than 0")
        self._tau = tau
        self._exitage = self._calc_exitage()



import numpy as np
from rtdpy.rtd import RTD, RTDInputError


class AD_oo(RTD):
    """
    Create Axial Dispersion with open-open boundary conditions
    Residence Time Distribution (RTD) model. [1]_

    .. math::

        E(t) = \\frac{\\sqrt{Pe}}{\\tau\\sqrt{4\\pi\\theta}}
        \\text{exp}\\left[\\frac{-Pe \\left(1-\\theta\\right)^2}
        {4\\theta}\\right]\\\\
        \\theta = t/\\tau

    Parameters
    ----------
    tau : scalar
        L/U, not the mean residence time, see tau_oo
    peclet : scalar
        Reactor Peclet number (L*U/D)
    dt : scalar
        Time step for RTD.
        ``dt>0``
    time_end : scalar
        End time for RTD.
        ``time_end>0``

    References
    ----------
    .. [1] Levenspiel O., Smith W.K. (1957) Notes on the diffusion-type model
           for the longitudinal mixing of fluids in flow.
           "Chemical Engineering Science", 6, 227-233

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import rtdpy
    >>> for pe in [10, 100]:
    ...     a = rtdpy.AD_oo(tau=1, peclet=pe, dt=.01, time_end=3)
    ...     plt.plot(a.time, a.exitage, label="peclet={}".format(pe))
    >>> plt.xlabel('Time')
    >>> plt.ylabel('Exit Age Function')
    >>> plt.legend()
    >>> plt.show()
    """

    def __init__(self, tau, peclet, dt, time_end):
        super().__init__(dt, time_end)

        self._peclet = None
        self._theta = None
        self._tau_oo = None
        self._tau = None

        self.tau = tau
        self.peclet = peclet

    def _calc_exitage(self):
        try:
            theta_safe = np.clip(self.theta, np.finfo(float).eps, None)
            output = 1 / self.tau \
                / np.sqrt(4. * np.pi * theta_safe / self.peclet) \
                * np.exp(-(1. - theta_safe) ** 2
                         / (4. * theta_safe / self.peclet))

        except AttributeError:
            output = None
        except TypeError:
            output = None
        return output

    def _calc_theta(self):
        try:
            output = self.time / self.tau
        except (AttributeError, TypeError):
            output = None
        return output

    def _calc_tau_oo(self):
        try:
            output = (1. + 2. / self.peclet) * self.tau
        except TypeError:
            output = None
        return output

    @property
    def theta(self):
        return self._theta

    @property
    def peclet(self):
        return self._peclet

    @peclet.setter
    def peclet(self, peclet):
        if peclet <= 0:
            raise RTDInputError('peclet less than zero')
        self._peclet = peclet
        self._tau_oo = self._calc_tau_oo()
        self._theta = self._calc_theta()
        self._exitage = self._calc_exitage()

    @property
    def tau(self):
        return self._tau

    @tau.setter
    def tau(self, tau):
        if tau <= 0:
            raise RTDInputError('tau less than zero')
        self._tau = tau
        self._tau_oo = self._calc_tau_oo()
        self._theta = self._calc_theta()
        self._exitage = self._calc_exitage()

    @property
    def tau_oo(self):
        """
        Mean Residence Time for open-open system is different than L/U
        """
        return self._tau_oo



"""Closed-closed axial dispersion model"""
import numpy as np
from scipy import integrate

from rtdpy.rtd import RTD, RTDInputError
from rtdpy.ad_cc_funcs import dudt, jac


class AD_cc(RTD):
    """
    Create Axial Dispersion with closed-closed boundary conditions
    Residence Time Distribution (RTD) model. [1]_ [2]_

    Solution of equation

    .. math::

        \\frac{\\partial C}{\\partial \\theta}
        = \\frac{1}{Pe}\\frac{\\partial^2 C}{\\partial z^2}
        - \\frac{\\partial C}{\\partial z}

    where :math:`\\theta = t/\\tau` is dimensionless time,
    :math:`z` is dimensionless length, and an impulse input at z=0
    with Danckwerts BCs

    .. math::

        E(t) = C(z=1, t)\\\\
        C_{in} = \\delta(t)\\\\
        C_{in} = C\\rvert_{z=0}
        - D\\frac{\\partial C}{\\partial z}\\biggr\\rvert_{z=0}\\\\
        \\frac{\\partial C}{\\partial z} = 0, z=1

    and initial conditions

    .. math::

        C=0 \\text{ for } t=0

    Parameters
    ----------
    tau : scalar
        L/U or mean residence time
    peclet : scalar
        Reactor Peclet number (L*U/D)
    dt : scalar
        Time step for RTD.
        ``dt>0``
    time_end : scalar
        End time for RTD.
        ``time_end>0``

    Other Parameters
    ----------------
    nx : optional
        Number of points to discretize 1D PDE
    a : optional
        Rate at which to introduce material.
        The inverse of a is the approximate amount of time to resolve the
        impulse input
    rtol : optional
        Relative tolerance to use in ODE solver
    atol : optional
        Absolute tolerance to use in ODE solver
    max_step : optional
        Maximum time step size (dimensionless) to use in ODE solver

    References
    ----------
    .. [1] Pearson J.R.A. (1959) A note on the "Danckwerts" boundary conditions
           for continuous flow reactors. "Chemical Engineering Science", 6,
           281-284.
    .. [2] Danckwerts P.V. (1953) Continuous flow systems: Distribution of Residence
           Times. "Chemical Engineering Science", 2, 1-13.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import rtdpy
    >>> for pe in [10, 100]:
    ...     a = rtdpy.AD_cc(tau=1, peclet=pe, dt=.01, time_end=3)
    ...     plt.plot(a.time, a.exitage, label='peclet={}'.format(pe))
    >>> plt.xlabel('Time')
    >>> plt.ylabel('Exit Age Function')
    >>> plt.legend()
    >>> plt.show()
    """

    def __init__(self, tau, peclet, dt, time_end, nx=200, a=10000, rtol=1e-5,
                 atol=1e-10, max_step=0.01):
        """init AD model"""
        super().__init__(dt, time_end)

        self._tau = None
        self._peclet = None
        self._pde_result = None

        self.nx = nx
        self.a = a
        self.rtol = rtol
        self.atol = atol
        self.max_step = max_step

        self.tau = tau
        self.peclet = peclet

    def _calc_pde(self):
        """Calculates the dimensionless pde result"""
        t_max = 10
        t_span = (0, t_max)

        x = np.linspace(0, 1, self.nx)

        u0 = np.zeros(self.nx)

        h = x[1] - x[0]

        j = jac(0, 0, self.peclet, h, self.nx, self.a)

        sol = integrate.solve_ivp(
            fun=lambda t, y: dudt(t, y, self.peclet, h, self.nx, self.a),
            jac=j, t_span=t_span, y0=u0, method="BDF", dense_output=True,
            rtol=self.rtol, atol=self.atol, max_step=self.max_step)

        return sol

    def _calc_exitage(self):
        """calculates exitage function"""
        try:

            output = self._pde_result.sol(self.time / self.tau)[-1] / self.tau
        except AttributeError:
            output = None
        return output

    @property
    def peclet(self):
        """Returns Peclet number"""
        return self._peclet

    @peclet.setter
    def peclet(self, peclet):
        """Set Peclet number"""
        if peclet <= 0:
            raise RTDInputError('peclet less than zero')
        self._peclet = peclet
        self._pde_result = self._calc_pde()
        self._exitage = self._calc_exitage()

    @property
    def tau(self):
        """Returns tau"""
        return self._tau

    @tau.setter
    def tau(self, tau):
        """Set tau"""
        if tau <= 0:
            raise RTDInputError('tau less than zero')
        self._tau = tau
        self._exitage = self._calc_exitage()



"""Define PDE for closed-closed Axial Dispersion model
"""
import numpy as np


def dudt(t, u, pe, h, n, a):
    """
    Define PDE equation for AD closed-closed
    """
    tmp = np.empty_like(u)
    for j in range(1, n - 1):
        tmp[j] = 1 / pe / h ** 2 * (u[j + 1] - 2 * u[j] + u[j - 1])\
            - 1 / (2 * h) * (u[j + 1] - u[j - 1])

    u0 = pe * 2 * h * a * np.exp(-a * t) + u[1] - pe * 2 * h * u[0]

    tmp[0] = 1 / pe / h ** 2 * (u[1] - 2 * u[0] + u0)\
        - 1 / (2 * h) * (u[1] - u0)

    tmp[n - 1] = 2 / pe / h ** 2 * (u[n - 2] - u[n - 1])
    return tmp


def jac(t, u, pe, h, n, a):
    """
    Define Jacobian for AD closed-closed
    """
    tmp = np.zeros([n, n])

    for i in range(n - 2):
        j = i + 1
        tmp[j, j - 1] = 1 / pe / h ** 2 + 1 / (2 * h)
        tmp[j, j] = -2 / pe / h ** 2
        tmp[j, j + 1] = 1 / pe / h ** 2 - 1 / (2 * h)

    tmp[0, 0] = 1 / pe / h ** 2 * (-2 - pe * 2 * h) - pe
    tmp[0, 1] = 2 / pe / h ** 2
    tmp[n - 1, n - 1] = -2 / pe / h ** 2
    tmp[n - 1, n - 2] = 2 / pe / h ** 2
    return tmp



import numpy as np

from rtdpy.rtd import RTD, RTDInputError


class Elist(RTD):
    """
    Create RTD model from a list of Rtd objects.  Combinations of models
    are convolved together. [1]_

    Parameters
    ----------
        elist : list
            List of Rtd objects.
            All objects must have same `dt` and `time_end`.

    References
    ----------
    .. [1] Levenspiel O. (1999) "Chemical Reaction Engineering: Third Edition"
           John Wiley & Sons, Inc.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import rtdpy
    >>> a = rtdpy.Ncstr(tau=1, n=1, dt=.01, time_end=15)
    >>> b = rtdpy.Pfr(tau=10, dt=.01, time_end=15)
    >>> c = rtdpy.Elist([a, b])
    >>> plt.plot(a.time, a.exitage, label="CSTR")
    >>> plt.plot(b.time, b.exitage, label="PFR")
    >>> plt.ylim(0, 1.1)
    >>> plt.title('Original RTD models')
    >>> plt.xlabel('Time')
    >>> plt.ylabel('Exit Age Function')
    >>> plt.legend()
    >>> plt.figure()
    >>> plt.plot(c.time, c.exitage)
    >>> plt.xlabel('Time')
    >>> plt.ylabel('Exit Age Function')
    >>> plt.title('Combination of models')
    >>> plt.show()
    """

    def __init__(self, elist):
        self._index = 0
        self._elist = None

        dt = self.get_elist_dt(elist)

        time_end = self.get_elist_time_end(elist)

        super().__init__(dt=dt, time_end=time_end)

        self.elist = elist

    @property
    def elist(self):
        return self._elist

    @elist.setter
    def elist(self, elist):

        self._elist = elist

        exitage = self.elist[0].exitage
        for idx, item in enumerate(self._elist[1:]):
            exitage = np.convolve(item.exitage, exitage) * self.dt
            exitage = exitage[:len(self.time) or None]
        self._exitage = exitage

    @staticmethod
    def get_elist_dt(elist):

        if any([not isinstance(item, RTD) for item in elist]):
            raise RTDInputError('item in list not an RTD class')

        dts = [item.dt for item in elist]
        if len(set(dts)) == 1:
            dt = dts[0]
        else:
            raise RTDInputError('RTDs have inconsistent dt in Elist')
        return dt

    @staticmethod
    def get_elist_time_end(elist):

        if any([not isinstance(item, RTD) for item in elist]):
            raise RTDInputError('item in list not an RTD class')

        time_ends = [item.time_end for item in elist]
        if len(set(time_ends)) == 1:
            time_end = time_ends[0]
        else:
            raise RTDInputError('RTDs have inconsistent dt in Elist')
        return time_end

    def __iter__(self):
        self._index = 0
        return self

    def __next__(self):
        try:
            result = self.elist[self._index]
        except IndexError:
            raise StopIteration
        self._index += 1
        return result



import numpy as np

from rtdpy.rtd import RTD


class Zusatz(RTD):
    """
    Create Zusatz Residence Time Distribution (RTD) model. [1]_
    Parameter a is chosen such that the integral is 1.

    .. math::

        E(t) = a t^{-c-1} b^{c+1}
        \\text{exp}\\left[\\left(b^c t^{-c} -1\\right)\\frac{-c-1}{c}\\right]
        \\\\a = \\frac{1+c}{b\\, \\text{exp}\\left[1+1/c\\right]}

    Parameters
    ----------
    b : scalar
        b Zusatz parameter
    c : scalar
        c Zusatz parameter
    dt : scalar
        Time step for RTD. ``dt>0``
    time_end : scalar
        End time for RTD. ``time_end>0``

    References
    ----------
    .. [1] Poulesquen A., et al. (2003) A study of residence time distribution
           in co-rotating twin-screw extruders.  Part II: Experimental
           validation. "Polymer Engineering and Science", 43(12), 1849-1862.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import rtdpy
    >>> for c in [3, 7]:
    >>>     a = rtdpy.Zusatz(b=25, c=c, dt=.01, time_end=100)
    >>>     plt.plot(a.time, a.exitage, label=f"c={c}")
    >>> plt.xlabel('Time')
    >>> plt.ylabel('Exit Age Function')
    >>> plt.legend()
    >>> plt.show()
    """
    def __init__(self, b, c, dt, time_end):
        super().__init__(dt, time_end)

        self._a = None
        self._b = None
        self._c = None

        self.b = b
        self.c = c

    def _calc_exitage(self):
        """equation for exit age function"""
        try:
            time_safe = np.clip(self.time, np.finfo(float).eps, None)
            output = self.a * time_safe**(-self.c-1) \
                * self.b**(self.c + 1) \
                * np.exp((self.b**self.c * time_safe**(-1 * self.c) - 1)
                         * (-self.c - 1) / self.c)
        except TypeError:
            output = None
        return output

    def _calc_a(self):
        """calculate a to normalize rtd"""
        try:
            output = (1 + self.c) / (self.b * np.exp(1 + 1 / self.c))
        except TypeError:
            output = None
        return output

    @property
    def a(self):
        """a parameter that normalizes RTD to 1"""
        return self._a

    @property
    def b(self):
        """b parameter"""
        return self._b

    @b.setter
    def b(self, b):
        # TODO: check for limits on input parameters
        self._b = b
        self._a = self._calc_a()
        self._exitage = self._calc_exitage()

    @property
    def c(self):
        """c parameter"""
        return self._c

    @c.setter
    def c(self, c):
        # TODO: check for limits on input parameters
        self._c = c
        self._a = self._calc_a()
        self._exitage = self._calc_exitage()



"""init file for rtdpy"""
from rtdpy.ncstr import Ncstr
from rtdpy.pfr import Pfr
from rtdpy.ad_oo import AD_oo
from rtdpy.elist import Elist
from rtdpy.zusatz import Zusatz
from rtdpy.ad_cc import AD_cc
from rtdpy.rtd import RTDInputError

__all__ = ['Ncstr',
           'Pfr',
           'AD_oo',
           'Elist',
           'Zusatz',
           'AD_cc',
           'RTDInputError',
           ]



from setuptools import setup

from rtdpy.const import version

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='rtdpy',
    version=version,
    packages=['rtdpy'],
    license='LICENSE',
    author='Matthew Flamm',
    author_email='matthew.flamm@merck.com',
    url='https://merck.github.io/rtdpy',
    description='Python package for residence time distribution analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['numpy',
                      'scipy'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering",
        ],
)
